/*
 * Copyright (C) 2006 Jason Smith
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/*!
 * @file defs.h
 * @brief Declarations of constants, flag, data structures, etc.
 */

#ifndef IGEMS_DEFS_H
#define IGEMS_DEFS_H

#include <pz.h>

#include "log.h"

/////////////////////////////////////////////////////////////////////
// utility-type definitions

#ifndef TRUE
#define TRUE   1
#endif

#ifndef FALSE
#define FALSE  0
#endif

#define SET_BIT( var, bit )       ( (var) |=  (bit) )
#define CLEAR_BIT( var, bit )     ( (var) &= ~(bit) )
#define IS_BIT_SET( var, bit )    ( (var) &   (bit) )

/////////////////////////////////////////////////////////////////////
// various defines for the game

/*! Title for the application window. */
#define IGEMS_TITLE         "iGems"

/*! Location in the pz2 menu to place our game. */
#define IGEMS_MENU_PATH      "/Extra/Games/iGems"

/*! pz2 menu ID for picking which skin is used. */
#define IGEMS_SKIN_OPTION    1
#define IGEMS_CURSOR_OPTION  2

/*!
 * Determines whether the game runs in fullscreen mode or windows.
 * Set to 1 for fullscreen mode.  Set to 0 for windowed.
 */
#define IGEMS_FULLSCREEN     1

/*! Milliseconds between calls to the main loop. */
#define IGEMS_LOOP_TIME      (50)

#define IGEMS_SWAP_TIME      (500/IGEMS_LOOP_TIME)   /*!< Number of frames (ms / frame time) to show a failed swap. */
#define IGEMS_MATCH_TIME     (500/IGEMS_LOOP_TIME)   /*!< Number of frames (ms / frame time) to show matches blocks. */
#define IGEMS_CLEAN_TIME     (500/IGEMS_LOOP_TIME)   /*!< Number of frames (ms / frame time) to remove matched blocks. */
#define IGEMS_COLLAPSE_TIME  (500/IGEMS_LOOP_TIME)   /*!< Number of frames (ms / frame time) to collapse the field. */

/////////////////////////////////////////////////////////////////////

/*! ID for an empty spot in the field. */
#define IGEMS_NO_BLOCK     0

/*! Number of block types */
#define IGEMS_NUM_BLOCKS   7

/*! The size, in blocks, of the field.  This is the number of rows and the number of columns. */
#define IGEMS_FIELD_SIZE   8

/*! Minimum number of blocks of the same type in a row required for a match. */
#define IGEMS_MATCH_SIZE   3

// definitions of the cursor direction, meaning should be obvious...
#define IGEMS_CURSOR_UP     0
#define IGEMS_CURSOR_RIGHT  1
#define IGEMS_CURSOR_DOWN   2
#define IGEMS_CURSOR_LEFT   3

/*! line width for the cursor drawing */
#define IGEMS_CURSOR_WIDTH  2

// state flags for GAME_DATA::flags
#define IGEMS_EXIT_BIT                0x01  /*!< Flag indicating the game should exit as soon as cleanly possible. */
#define IGEMS_CHECK_MATCHES_BIT       0x02  /*!< Indicates the game should check the field for matches.  See igems_CheckMatches(). */
#define IGEMS_CLEAN_FIELD_BIT         0x04  /*!< Indicates that all blocks marked to kill should be removed from the field.  See igems_CleanField(). */
#define IGEMS_COLLAPSE_FIELD_BIT      0x08  /*!< Indicates gravity should be applied to the field, moving all blocks to their lowest position.  See igems_CollapseField(). */
#define IGEMS_FILL_FIELD_BIT          0x10  /*!< Indicates any empty/invalid blocks in the field should be replaced with a new random block.  See igems_FillField(). */
#define IGEMS_SWAP_BLOCKS_BIT         0x20  /*!< Indicates that an attempt to swap adjacent blocks has been made. */
#define IGEMS_NO_MATCHES_BIT          0x40  /*!< Indicates there are no more near/possible matches. */

/*! Superset of all flags which may indicate the game is in a busy state (i.e. animating something to the user). */
#define IGEMS_GAME_BUSY_BITS   (IGEMS_CHECK_MATCHES_BIT \
								| IGEMS_CLEAN_FIELD_BIT \
								| IGEMS_COLLAPSE_FIELD_BIT \
								| IGEMS_FILL_FIELD_BIT \
								| IGEMS_SWAP_BLOCKS_BIT \
								)

#define IGEMS_IS_GAME_BUSY( var )    IS_BIT_SET( var, IGEMS_GAME_BUSY_BITS )

///////////////////////////////////////////////////////////////////////////////

/*!
 * @typedef IGEMS_BLOCK
 * @brief Data structure for a gem/block in our game.
 */
typedef struct _IGEMS_BLOCK
{
	int   type;     /*!< ID from 0 to IGEMS_NUM_BLOCKS indicating the type of block. */
	int   kill;     /*!< Flag to indicated whether this block is to be destroyed (e.g. because it was part of a match). */
	int   animate;  /*!< Data used by the renderer to animate various effects. */
} IGEMS_BLOCK;


/*!
 * @typedef IGEMS_CURSOR
 * The user's cursor, used to move blocks.
 */
typedef struct _IGEMS_CURSOR
{
	int x;         /*!< The column of the current cursor location. */
	int y;         /*!< The row of the current cursor location. */
	int pressed;   /*!< Flag indicating whether the cursor is 'pressed' (1) or not (0). */
	int direction; /*!< If (pressed == 1), this determines the direction the user is currently aiming for switching blocks. */
} IGEMS_CURSOR;


/*!
 * @typedef IGEMS_SWAPINFO
 * @brief Information on an attempted block swap.
 */
typedef struct _IGEMS_SWAPINFO
{
	IGEMS_BLOCK*  b1;      /*!< The block the user originally pressed the cursor on. */
	IGEMS_BLOCK*  b2;      /*!< The block choosen to be swapped with b1. */
	int animate;          /*!< Data for the renderer to animate the swapping. */
	int isOK;             /*!< Flag indicating whether the swap being processed is the inital
						   *  attempt at a swap (TRUE), or if the initial swap failed (because it did
						   *  not generate any matches) and so the current 'swap' is resetting
						   *  the blocks to their original positions (FALSE).
						   */
} IGEMS_SWAPINFO;


/*!
 * @typedef IGEMS_GAMEDATA
 * @brief Superset of all data required to run the game.
 */
typedef struct _IGEMS_GAMEDATA
{
	IGEMS_BLOCK*    field;     /*!< Pointer to the game field.  This data is dynamically allocated when the game starts */
	IGEMS_CURSOR    cursor;    /*!< The cursor used by the user. */
	IGEMS_SWAPINFO  swapInfo;  /*!< Information on the current block swap.  Only valid if the IGEMS_SWAP_BLOCKS_BIT is set for flags. */

	int  needToWait;           /*!< Flag indicating that an event occurred requiring the game logic to pause briefly before continueing. */
	int  framesToWait;         /*!< The number of frame updates to wait before restarting game logic. */

	int  flags;                /*!< Various state flags for the game. */

	int  dirtyCols;            /*!< Columns that are dirty and need rendered. */

} IGEMS_GAMEDATA;


/*!
 * @typedef IGEMS_GLOBALS
 * @brief Container for all of the global variables.
 * @todo Document members, remove if possible.
 */
typedef struct _IGEMS_GLOBALS
{
	PzModule*   module;
	PzWindow*   window;
	PzWidget*   widget;
	PzConfig*   config;

	IGEMS_GAMEDATA   game;   /*!< Actual game data. */



} IGEMS_GLOBALS;

extern IGEMS_GLOBALS  g_globals;

#define MARK_DIRTY_COLS( col )  SET_BIT( g_globals.game.dirtyCols, (1<<(col+1)) )
#define IS_COL_DIRTY( col )     IS_BIT_SET( g_globals.game.dirtyCols, (1<<(col+1)) )
#define MARK_CLEAN_COL( col )   CLEAR_BIT( g_globals.game.dirtyCols, (1<<(col+1)) )

#endif // IGEMS_DEFS_H

