/*
 * Copyright (C) 2006 Jason Smith
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/*!
 * @file game.c
 * @brief Implementation of functions to handle game manipulation and logic.
 */

#include <pz.h>
#include <string.h>

#include "log.h"
#include "defs.h"
#include "game.h"
#include "util.h"

/////////////////////////////////////////////////////////////////////
void igems_InitBlock( IGEMS_BLOCK* block )
{
	JFS_ASSERT( block );
	if( block )
	{
		block->type = 1 + igems_Rand( IGEMS_NUM_BLOCKS );
		block->kill = FALSE;
		block->animate = 0;
	}
}

/////////////////////////////////////////////////////////////////////
void igems_CopyBlock( IGEMS_BLOCK* source, IGEMS_BLOCK* dest )
{
	JFS_ASSERT( source );
	JFS_ASSERT( dest );

	if( source && dest && (source != dest) )
	{
		memcpy( (void*) dest, (void*) source, sizeof(IGEMS_BLOCK) );
	}
}

/////////////////////////////////////////////////////////////////////
IGEMS_BLOCK* igems_CreateField( void )
{
	IGEMS_BLOCK* field = (IGEMS_BLOCK*) malloc( sizeof(IGEMS_BLOCK) * IGEMS_FIELD_SIZE * IGEMS_FIELD_SIZE );
	if( !field )
	{
		ERROR_LOG( "Failed to create game field, out of memory!!\n" );
		return NULL;
	}

	DEBUG_LOG( "Successfully created game field at 0x%08x\n", (unsigned int)(field) );
	return field;
}

/////////////////////////////////////////////////////////////////////
void igems_DestroyField( IGEMS_BLOCK* field )
{
	// dont assert, allow this to be called at any time
	if( field )
	{
		DEBUG_LOG( "Deleting field at 0x%08x\n", (unsigned int)(field) );
		free( field );
	}
}

/////////////////////////////////////////////////////////////////////
void igems_InitField( IGEMS_BLOCK* field )
{
	JFS_ASSERT( field );
	if( field )
	{
		int i;

		// go through and initialize every block in the field
		for( i=0; i<(IGEMS_FIELD_SIZE*IGEMS_FIELD_SIZE); i++ )
		{
			igems_InitBlock( &(field[i]) );
		}
	}
}

/////////////////////////////////////////////////////////////////////
IGEMS_BLOCK* igems_GetBlock( IGEMS_BLOCK* field, int col, int row )
{
	JFS_ASSERT( field );
	JFS_ASSERT( (col>=0 && col<IGEMS_FIELD_SIZE) );
	JFS_ASSERT( (row>=0 && row<IGEMS_FIELD_SIZE) );

	if( !field || (col<0) || (row<0) || (col>=IGEMS_FIELD_SIZE) || (row>=IGEMS_FIELD_SIZE) )
	{
		ERROR_LOG( "Invalid parameter(s) to igems_GetBlock()\n" );
		return NULL;
	}

	return &( field[(row * IGEMS_FIELD_SIZE) + col] );
}

/////////////////////////////////////////////////////////////////////
int igems_GetBlockType( IGEMS_BLOCK* field, int col, int row )
{
	if( !field || (col<0) || (row<0) || (col>=IGEMS_FIELD_SIZE) || (row>=IGEMS_FIELD_SIZE) )
	{
		return IGEMS_NO_BLOCK;
	}

	return (field[ (row * IGEMS_FIELD_SIZE) + col ]).type;
}

/////////////////////////////////////////////////////////////////////
void igems_InitCursor( IGEMS_CURSOR* cursor )
{
	JFS_ASSERT( cursor );
	if( cursor )
	{
		cursor->x          = IGEMS_FIELD_SIZE/2;
		cursor->y          = IGEMS_FIELD_SIZE/2;
		cursor->pressed    = FALSE;
		cursor->direction  = IGEMS_CURSOR_UP;
	}
}

/////////////////////////////////////////////////////////////////////
int igems_ScrollCursor( IGEMS_GAMEDATA* game, int numMoves )
{
	JFS_ASSERT( game );
	JFS_ASSERT( game->field );

	if( numMoves == 0 )
	{
		// there are no moves, so just return.
		return TTK_EV_UNUSED;
	}

	if( game->cursor.pressed )
	{
		int d;
		int fixed;

		DEBUG_LOG( "scroll_cursor: cursor is pressed, changing swap direction.\n" );

		if( game->cursor.direction == IGEMS_CURSOR_LEFT )
			MARK_DIRTY_COLS( game->cursor.x - 1 );
		if( game->cursor.direction == IGEMS_CURSOR_RIGHT )
			MARK_DIRTY_COLS( game->cursor.x + 1 );

		// if the cursor is pressed, then we need to rotate our swap direction
		// there are 4 different swap directions (up, right, down, left), so
		// count each direction as one move
		game->cursor.direction += numMoves;
		if( numMoves > 0 )
		{
			d = 1;
			// make sure we're in the range of our 4 directions
			game->cursor.direction %= 4;
		}
		else
		{
			d = (-1);
			// make sure we're in the range of our 4 directions
			while( game->cursor.direction < 0 )
				game->cursor.direction += 4;
		}

		// if we're on the edge of the field, the swap may be pointing at nothing
		// fix this by making it point to something meaningful.

		do
		{
			fixed = FALSE;

			if( (game->cursor.x == 0) && (game->cursor.direction == IGEMS_CURSOR_LEFT) )
			{
				game->cursor.direction += d;
				fixed = TRUE;
			}
			else if( (game->cursor.x == (IGEMS_FIELD_SIZE-1)) && (game->cursor.direction == IGEMS_CURSOR_RIGHT) )
			{
				game->cursor.direction += d;
				fixed = TRUE;
			}

			if( (game->cursor.y == 0) && (game->cursor.direction == IGEMS_CURSOR_UP) )
			{
				game->cursor.direction += d;
				fixed = TRUE;
			}
			else if( (game->cursor.y == (IGEMS_FIELD_SIZE-1)) && (game->cursor.direction == IGEMS_CURSOR_DOWN) )
			{
				game->cursor.direction += d;
				fixed = TRUE;
			}

			// make sure the direction is valid.
			while( game->cursor.direction < 0 )
			{
				game->cursor.direction += 4;
				fixed = TRUE;
			}

			if( game->cursor.direction >=4 )
			{
				game->cursor.direction %= 4;
				fixed = TRUE;
			}

		} while( fixed );

		MARK_DIRTY_COLS( game->cursor.x );
		if( game->cursor.direction == IGEMS_CURSOR_LEFT )
			MARK_DIRTY_COLS( game->cursor.x - 1 );
		else if( game->cursor.direction == IGEMS_CURSOR_RIGHT )
			MARK_DIRTY_COLS( game->cursor.x + 1 );
	}
	else // cursor is not currently pressed
	{
		// mark starting column as dirty
		MARK_DIRTY_COLS( game->cursor.x );

		// determine our current location, in terms of index in the game->field array
		int curLocation = game->cursor.x + (game->cursor.y * IGEMS_FIELD_SIZE);

		// cursor scrolls will move the cursor by a some number of blocks to the
		// left or right in the field array.  this computes our new location
		// in that 1-D array.
		int newLocation = curLocation + numMoves;

		DEBUG_LOG( "scroll_cursor: not pressed, moving cursor\n" );

		// check for underflow
		while( newLocation < 0 )
		{
			newLocation += (IGEMS_FIELD_SIZE*IGEMS_FIELD_SIZE);
		}

		// check for overflow
		while( newLocation >= (IGEMS_FIELD_SIZE*IGEMS_FIELD_SIZE) )
		{
			newLocation -= (IGEMS_FIELD_SIZE*IGEMS_FIELD_SIZE);
		}

		// convert back to (x,y) format.
		game->cursor.x = newLocation % IGEMS_FIELD_SIZE;
		game->cursor.y = newLocation / IGEMS_FIELD_SIZE;

		MARK_DIRTY_COLS( game->cursor.x ); // mark ending column dirty
	}

	return TTK_EV_CLICK;
}

/////////////////////////////////////////////////////////////////////
void igems_SelectCursor( IGEMS_GAMEDATA* game )
{
	JFS_ASSERT( game );
	JFS_ASSERT( game->field );

	if( game->cursor.pressed )
	{
		int x, y;
		IGEMS_BLOCK* b1;
		IGEMS_BLOCK* b2;

		x = game->cursor.x;
		y = game->cursor.y;

		// get the block the cursor is actually on
		b1 = igems_GetBlock( game->field, x, y );

		// adjust x,y based on the cursor direction
		switch( game->cursor.direction )
		{
			case IGEMS_CURSOR_UP:    --y;   break;
			case IGEMS_CURSOR_RIGHT: ++x;   break;
			case IGEMS_CURSOR_DOWN:  ++y;   break;
			case IGEMS_CURSOR_LEFT:  --x;   break;
			default:
				ERROR_LOG( "Invalid cursor direction in press_cursor(): direciton=%i\n", game->cursor.direction );
				JFS_ASSERT( 0 );
				return;
		} // switch

		// now get the block that the cursor wants to swap with.
		b2 = igems_GetBlock( game->field, x, y );

		JFS_ASSERT( b1 );
		JFS_ASSERT( b2 );
		if( !(b1 && b2) )
		{
			ERROR_LOG( "Invalid swap blocks in press_cursor().  Will attempt to recover.\n" );
			game->cursor.pressed = FALSE;
			return;
		}

		MARK_DIRTY_COLS( game->cursor.x );
		if( game->cursor.direction == IGEMS_CURSOR_LEFT )
			MARK_DIRTY_COLS( game->cursor.x - 1 );
		else if( game->cursor.direction == IGEMS_CURSOR_RIGHT )
			MARK_DIRTY_COLS( game->cursor.x + 1 );

		// set the flag for this event..
		SET_BIT( game->flags, IGEMS_SWAP_BLOCKS_BIT );
		game->swapInfo.b1 = b1;
		game->swapInfo.b2 = b2;
		game->swapInfo.animate = 0;
		game->swapInfo.isOK = TRUE; // until proven otherwise
	}
	else // not pressed
	{
		DEBUG_LOG( "press_cursor: setting cursor to pressed state.\n" );
		game->cursor.pressed = TRUE;

		// check we're not swapping with nothing...
		if( (game->cursor.x == 0) && (game->cursor.direction==IGEMS_CURSOR_LEFT) )
			game->cursor.direction = IGEMS_CURSOR_RIGHT;

		if( (game->cursor.x == IGEMS_FIELD_SIZE-1) && (game->cursor.direction==IGEMS_CURSOR_RIGHT) )
			game->cursor.direction = IGEMS_CURSOR_LEFT;

		if( (game->cursor.y == 0) && (game->cursor.direction==IGEMS_CURSOR_UP) )
			game->cursor.direction = IGEMS_CURSOR_DOWN;

		if( (game->cursor.y == IGEMS_FIELD_SIZE-1) && (game->cursor.direction==IGEMS_CURSOR_DOWN) )
			game->cursor.direction = IGEMS_CURSOR_UP;

		MARK_DIRTY_COLS( game->cursor.x );
		if( game->cursor.direction == IGEMS_CURSOR_LEFT )
			MARK_DIRTY_COLS( game->cursor.x - 1 );
		else if( game->cursor.direction == IGEMS_CURSOR_RIGHT )
			MARK_DIRTY_COLS( game->cursor.x + 1 );
	}
}

/////////////////////////////////////////////////////////////////////
// this is a brute force 'check every block' method.
int igems_CheckForMatches( IGEMS_GAMEDATA* game )
{
	IGEMS_BLOCK* f;
	int x, y, i;

	int numFound;
	int currentType;
	int startPos;

	int foundMatch = FALSE;

	JFS_ASSERT( game );
	JFS_ASSERT( game->field );

	// start with looking at every row
	for( y=0; y<IGEMS_FIELD_SIZE; ++y )
	{
		// when looking at the rows, f is a pointer to the start of the row.
		f = &( game->field[ y*IGEMS_FIELD_SIZE ] );

		numFound = 0;
		currentType = IGEMS_NO_BLOCK;
		startPos = 0;

		// check every column in the row
		for( x=0; x<IGEMS_FIELD_SIZE; x++ )
		{
			if( (currentType == IGEMS_NO_BLOCK) || (currentType != f[x].type) )
			{
				// either we haven't started a match, or this block is not part
				// of the previous match.  Start a new match ...
				currentType = f[x].type;
				startPos    = x;
				numFound    = 1;
			}
			else // part of a match
			{
				++numFound;

				if( numFound == IGEMS_MATCH_SIZE )
				{
					foundMatch = TRUE;
					// kill each block in the match
					for( i=startPos; i<=x; i++ )
					{
						f[i].kill = TRUE;
						MARK_DIRTY_COLS( i );
					}
				}
				else if( numFound > IGEMS_MATCH_SIZE )
				{
					// already killed the other blocks in this match, kill this one too
					f[x].kill = TRUE;
					MARK_DIRTY_COLS( x );
				}
				else
				{
					// not enough for a match, so do nothing
				}
			} // else part of match
		} // for every column in the row
	} // for every row


	// now check every column in the field.
	for( x=0; x<IGEMS_FIELD_SIZE; x++ )
	{
		numFound = 0;
		currentType = IGEMS_NO_BLOCK;
		startPos = 0;

		// check every column in the row
		for( y=0; y<IGEMS_FIELD_SIZE; y++ )
		{
			// for the columns, f is the single block we are currently looking at.
			f = igems_GetBlock( game->field, x, y );

			if( (currentType == IGEMS_NO_BLOCK) || (currentType != f->type) )
			{
				// either we haven't started a match, or this block is not part
				// of the previous match.  Start a new match ...
				currentType = f->type;
				startPos    = y;
				numFound    = 1;
			}
			else // part of a match
			{
				++numFound;

				if( numFound == IGEMS_MATCH_SIZE )
				{
					foundMatch = TRUE;
					MARK_DIRTY_COLS( x ); // we're looking at columns, so only one mark needed.

					// kill each block in the match
					for( i=startPos; i<=y; i++ )
					{
						igems_GetBlock(game->field, x, i)->kill = TRUE;
					}
				}
				else if( numFound > IGEMS_MATCH_SIZE )
				{
					// already killed the other blocks in this match, kill this one too
					f->kill = TRUE;
				}
				else
				{
					// not enough for a match, so do nothing
				}
			} // else part of match
		} // for every column in the row
	} // for every row

	// we just checked for matches, make sure the game state flag is disabled.
	CLEAR_BIT( game->flags, IGEMS_CHECK_MATCHES_BIT );

	if( foundMatch )
	{
		SET_BIT( game->flags, IGEMS_CLEAN_FIELD_BIT );
	}
	return foundMatch;
}

/////////////////////////////////////////////////////////////////////
int igems_CleanField( IGEMS_GAMEDATA* game )
{
	int i;
	int numCleaned = 0;

	JFS_ASSERT( game );
	JFS_ASSERT( game->field );

	// we don't need to go throuh in rows or columns,
	// so go straight through the entire field.
	for( i=0; i<(IGEMS_FIELD_SIZE*IGEMS_FIELD_SIZE); i++ )
	{
		// check if it is marked to be killed
		if( game->field[i].kill )
		{
			// clean/delete this block
			game->field[i].type = IGEMS_NO_BLOCK;
			++numCleaned;

			MARK_DIRTY_COLS( i%IGEMS_FIELD_SIZE );

			DEBUG_LOG( "cleaning block %i (x,y)\n", i, (i%IGEMS_FIELD_SIZE), (i/IGEMS_FIELD_SIZE) );
		}
	}

	DEBUG_LOG( "clean_field: found %i blocks to clean.\n", numCleaned );
	CLEAR_BIT( game->flags, IGEMS_CLEAN_FIELD_BIT );
	if( numCleaned > 0 )
	{
		SET_BIT( game->flags, IGEMS_COLLAPSE_FIELD_BIT );
	}
	return numCleaned;
}

/////////////////////////////////////////////////////////////////////
void igems_CollapseField( IGEMS_GAMEDATA* game )
{
	int shift;
	int x, y;
	IGEMS_BLOCK* f;

	DEBUG_LOG( "%s()\n", __func__ );
	JFS_ASSERT( game );
	JFS_ASSERT( game->field );

	for( x=0; x<IGEMS_FIELD_SIZE; x++ )
	{
		shift = 0;

		// start at the bottom of each column and work our way up.
		for( y=IGEMS_FIELD_SIZE-1; y>=0; y-- )
		{
			f = igems_GetBlock( game->field, x, y );
			JFS_ASSERT( f );

			if( f->type == IGEMS_NO_BLOCK )
			{
				++shift;
			}
			else if( shift > 0 ) // we need to shift this block
			{
				IGEMS_BLOCK* f2 = igems_GetBlock( game->field, x, y+shift );
				JFS_ASSERT( f2 );

				// copy the block to it's new location
				igems_CopyBlock( f, f2 );

				// erase the original location
				f->type = IGEMS_NO_BLOCK;
			}
			// else shift == 0, so don't do anything

		} // for every row/y

		if( shift != 0 )
			MARK_DIRTY_COLS( x );
	} // for every column/x

	CLEAR_BIT( game->flags, IGEMS_COLLAPSE_FIELD_BIT );
	SET_BIT( game->flags, IGEMS_FILL_FIELD_BIT );
}

/////////////////////////////////////////////////////////////////////
int igems_FillField( IGEMS_GAMEDATA* game )
{
	int i;
	int numFilled = 0;

	JFS_ASSERT( game );
	JFS_ASSERT( game->field );

	// we don't need to go throuh in rows or columns,
	// just go straight through the entire field.
	for( i=0; i<(IGEMS_FIELD_SIZE*IGEMS_FIELD_SIZE); i++ )
	{
		// check if there is no block in this location
		if( game->field[i].type == IGEMS_NO_BLOCK )
		{
			igems_InitBlock( &(game->field[i]) );
			++numFilled;

			MARK_DIRTY_COLS( i%IGEMS_FIELD_SIZE );
		}
	}

	DEBUG_LOG( "fill_field: found %i blocks to fill\n", numFilled );

	CLEAR_BIT( game->flags, IGEMS_FILL_FIELD_BIT );
	if( numFilled > 0 )
	{
		SET_BIT( game->flags, IGEMS_CHECK_MATCHES_BIT );
	}
	return numFilled;
}


/////////////////////////////////////////////////////////////////////
/*
 *  All we need to do is look for a single possible match.  If we
 *  find at least one, then we can immediately return TRUE.

 *   x--  -x-  --x
 *   -XX  x-x  xx-
 */
int igems_CheckForNearMatches( IGEMS_GAMEDATA* game )
{
	int x, y, curType;
	int len = (IGEMS_FIELD_SIZE - 2);
	IGEMS_BLOCK* field;

	JFS_ASSERT( game );
	JFS_ASSERT( game->field );

	DEBUG_LOG( "%s()\n", __func__ );

	field = game->field;
	for( y=0; y<len; y++ )
	{
		for( x=0; x<len; x++ )
		{
			curType = igems_GetBlockType( game->field, x, y );
			if( curType == IGEMS_NO_BLOCK )
			{
				// ignore this empty block
				continue;
			}

			// check for 2 in a row horizontally ...
			if( curType == igems_GetBlockType( field, x+1, y ) )
			{
				if( curType == igems_GetBlockType( field, x+2, y-1 ) )
					return TRUE;

				if( curType == igems_GetBlockType( field, x+2, y+1 ) )
					return TRUE;

				if( curType == igems_GetBlockType( field, x+3, y ) )
					return TRUE;
			}

			// check for 2 in a row vertically ...
			if( curType == igems_GetBlockType( field, x, y+1 ) )
			{
				if( curType == igems_GetBlockType( field, x-1, y+2 ) )
					return TRUE;

				if( curType == igems_GetBlockType( field, x+1, y+2 ) )
					return TRUE;

				if( curType == igems_GetBlockType( field, x, y+3 ) )
					return TRUE;
			}

			// check for a 'hill' or 'valley' horizontally ...
			if( curType == igems_GetBlockType( field, x+2, y ) )
			{
				// 'hill'
				if( curType == igems_GetBlockType( field, x+1, y-1 ) )
					return TRUE;

				// 'valley'
				if( curType == igems_GetBlockType( field, x+1, y+1 ) )
					return TRUE;
			}

			// check for a 'hill' or 'valley' horizontally ...
			if( curType == igems_GetBlockType( field, x+2, y ) )
			{
				// 'hill'
				if( curType == igems_GetBlockType( field, x+1, y-1 ) )
					return TRUE;

				// 'valley'
				if( curType == igems_GetBlockType( field, x+1, y+1 ) )
					return TRUE;
			}

		} // for almost every column
	} // for almost every row

	DEBUG_LOG( "Failed to find any near matches.\n" );
	SET_BIT( game->flags, IGEMS_NO_MATCHES_BIT );
	return FALSE;
}

