/*
 * Copyright (C) 2006 Jason Smith
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/*!
 * @file game.h
 * @todo document this file.
 * @todo Add prefixes to macros/functions to split up the namespace.
 */

#ifndef GAME_H
#define GAME_H

#include "defs.h"

/*!
 * @brief Initialize the default/random values for a single block
 * @param block Pointer to the block to initialize.
 */
void igems_InitBlock( IGEMS_BLOCK* block );

/*!
 * @brief Copy one block over another.
 * @param source The block to copy.
 * @param dest The block to overwrite.
 */
void igems_CopyBlock( IGEMS_BLOCK* source, IGEMS_BLOCK* dest );

/*!
 * @brief Allocate a new game field.
 * @returns Returns a pointer to the new game field, or NULL if there was an error.
 */
IGEMS_BLOCK* igems_CreateField( void );

/*!
 * @brief Destory game field created by create_field().
 * @param field Pointer to the field to destroy.
 */
void igems_DestroyField( IGEMS_BLOCK* field );

/*!
 * @brief Initialize the game field.
 * @param field Pointer to the field (as returned by create_field).
 *
 * Initializes the game field by filling the board with random blocks.  In
 * our game, blocks are the same as gems.
 */
void igems_InitField( IGEMS_BLOCK* field );

/*!
 * @brief Get a pointer to the block at position (col,row) in the field.
 * @param field Pointer to the game field to retrieve the block from.
 * @param col The column in the field that the block is in.
 * @param row The row in the field that the block is in.
 * @returns Returns a pointer to the desired block, or NULL if there is
 *   an error such as an invalid row or column specified.
 */
IGEMS_BLOCK* igems_GetBlock( IGEMS_BLOCK* field, int col, int row );

/*! Initialize the specified cursor to the default values. */
void igems_InitCursor( IGEMS_CURSOR* cursor );

/*!
 * @brief Handle scrolling the cursor to the right or left.
 * @param game The game containing the cursor to move.
 * @param arg The arg field from the scroll event (i.e. how much the user scrolled).
 *
 * The IGEMS_GAMEDATA* is needed because we may need to call secondary events,
 * such as a block swap, etc.
 */
int igems_ScrollCursor( IGEMS_GAMEDATA* game, int arg );

/*!
 * @brief Handle the user 'pressing' the cursor.
 * @param game The game containing the cursor being pressed.
 *
 * Handles the user 'pressing' the cursor, in an attempt to start a swap or
 * to finish a swap.
 *
 * If the cursor is currently pressed, then this initiates a swap.  If not
 * pressed, this sets the cursor into the pressed state.
 */
void igems_SelectCursor( IGEMS_GAMEDATA* game );

/*!
 * @brief Search through the field and look for matches to clear (marking the blocks as appropriate).
 * @param game The game data to search.
 * @returns Returns non-zero if matches are found, or zero if no matches can be found.
 *
 * Scans the game field and looks for matches.  If matches are found, it marks the blocks in the match
 * to be killed, but otherwise leaves the board untouched.  It does not look at the 'kill' flag for
 * the blocks.
 */
int igems_CheckForMatches( IGEMS_GAMEDATA* game );

/*!
 * @brief Scans through the field and removes any blocks that are marked to be killed.
 * @param game The game containing the field to search through.
 * @returns Returns the number of blocks removed from the field.  This may be zero.
 */
int igems_CleanField( IGEMS_GAMEDATA* game );

/*!
 * @brief Applies gravity to each column so blocks fall to their lowest position.
 * @param game The game containing the field to collapse.
 *
 * After cleaning the field with clean_field(), there may be holes in the field.  This
 * function collapses all blocks towards the bottom of the field to make room for new blocks
 * to arrive from the sky.  Note that this function does _not_ add any new blocks to the field.
 */
void igems_CollapseField( IGEMS_GAMEDATA* game );

/*!
 * @brief Fill in the field with blocks so that there are no empty spaces.
 * @param game The game containing the field to fill.
 * @returns Returns the number of blocks inserted into the field.
 *
 * After cleaning and collapsing the field, there may be empty blocks at the top
 * of the field.  This function fills in the field so that every space has a valid block.
 */
int igems_FillField( IGEMS_GAMEDATA* game );

int igems_CheckForNearMatches( IGEMS_GAMEDATA* game );

#endif // GAME_H
