/*
 * Copyright (C) 2006 Jason Smith
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/*!
 * @file igems.c
 * Main source file for the iGems game.  Contains pz2 module initialization and
 * related TTK callbacks.
 */

#include <pz.h>
#include <time.h>
#include <string.h>

#include "log.h"
#include "defs.h"
#include "game.h"
#include "render.h"

///////////////////////////////////////////////////////////
/*! Global variables for our game. */
IGEMS_GLOBALS  g_globals;

///////////////////////////////////////////////////////////
/*! TTK callback for cleaning up resources when the game quits. */
void igems_Cleanup( void )
{
	DEBUG_LOG( "%s()\n", __func__ );

	if( g_globals.game.field )
	{
		igems_DestroyField( g_globals.game.field );
		g_globals.game.field = NULL;
	}

	igems_DeinitRenderer();

	DEBUG_LOG( ">>>>>>>>>>>>>>>>>>>>\n" );
}

///////////////////////////////////////////////////////////
int igems_GameLoop( PzWidget* wid )
{
	int dirty = 0;
	IGEMS_GAMEDATA* game = &(g_globals.game);

	if( IS_BIT_SET( game->flags, IGEMS_EXIT_BIT) )
	{
		INFO_LOG( "%s: got EXIT flag, exiting.\n", __func__ );
		pz_close_window( wid->win );
	}

	// check if the logic is on hold, and if so update the wait count.
	if( game->needToWait )
	{
		game->framesToWait--;
		if( game->framesToWait <= 0 )
		{
			game->needToWait = FALSE;
		}
	}

	// Here is the game logic...
	//
	while( IGEMS_IS_GAME_BUSY(game->flags) && !(game->needToWait) )
	{
		// check if swapping blocks ...
		if( IS_BIT_SET(game->flags, IGEMS_SWAP_BLOCKS_BIT) )
		{
			if( game->swapInfo.isOK )
			{
				// then we're to try to go ahead with the initial swap
				IGEMS_BLOCK temp;

				igems_CopyBlock( game->swapInfo.b1, &temp );
				igems_CopyBlock( game->swapInfo.b2, game->swapInfo.b1 );
				igems_CopyBlock( &temp, game->swapInfo.b2 );

				if( igems_CheckForMatches( game ) )
				{
					// there was at least one match so the swap is OK

					// wait a little bit to allow the user to see
					game->needToWait = TRUE;
					game->framesToWait = IGEMS_MATCH_TIME;

					// no longer swapping blocks.
					CLEAR_BIT( game->flags, IGEMS_SWAP_BLOCKS_BIT );
					DEBUG_LOG( "Swap OK, matches have been found.\n" );
					game->cursor.pressed = FALSE; // reset the cursor
					dirty = 1;  // needs repainted
				}
				else
				{
					// the swap failed, but wait a short amount of time so the user can see we tried to do the swap
					game->swapInfo.isOK = FALSE;
					game->needToWait = TRUE;
					game->framesToWait = IGEMS_SWAP_TIME;
					DEBUG_LOG( "having to revert swap, no matches found.\n" );
					dirty = 1;
				}
			}
			else // the swap has already failed,.. revert it...
			{
				IGEMS_BLOCK temp;
				DEBUG_LOG( "Reverting swapped blocks ..." );

				igems_CopyBlock( game->swapInfo.b1, &temp );
				igems_CopyBlock( game->swapInfo.b2, game->swapInfo.b1 );
				igems_CopyBlock( &temp, game->swapInfo.b2 );

				dirty = 1;
				game->cursor.pressed = FALSE;
				CLEAR_BIT( game->flags, IGEMS_SWAP_BLOCKS_BIT );
			}

			continue;
		} // swap blocks flag

		if( IS_BIT_SET( game->flags, IGEMS_CHECK_MATCHES_BIT ) )
		{
			if( igems_CheckForMatches( game ) )
			{
				// found matches, ... wait a few frames to allow the render to draw something
				game->needToWait = TRUE;
				game->framesToWait = IGEMS_MATCH_TIME;
				dirty = 1;
			}
			else if( !(game->flags & (IGEMS_FILL_FIELD_BIT | IGEMS_CLEAN_FIELD_BIT | IGEMS_COLLAPSE_FIELD_BIT)) )
			{
				// we didn't find any matches, so make sure there are possible matches in the field
				if( !igems_CheckForNearMatches( game ) )
				{
					pz_message( "There are no matches left!!" );

//					SET_BIT( game->flags, IGEMS_EXIT_BIT );
//					INFO_LOG( "Closing window because there are no more matches left in the field.\n" );
//					pz_close_window( wid->win );
				}
			}
			continue;
		}

		if( IS_BIT_SET( game->flags, IGEMS_CLEAN_FIELD_BIT ) )
		{
			if( igems_CleanField( game ) )
			{
				// cleaned blocks from the field...
				//game->needToWait = TRUE;
				//game->framesToWait = (TIME_CLEANED / JFS_LOOP_TIME);
				dirty = 1;
			}
			continue;
		}

		if( IS_BIT_SET( game->flags, IGEMS_COLLAPSE_FIELD_BIT ) )
		{
			igems_CollapseField( game );
			game->needToWait = TRUE;
			game->framesToWait = IGEMS_COLLAPSE_TIME;

			dirty =1;
			continue;
		}

		if( IS_BIT_SET( game->flags, IGEMS_FILL_FIELD_BIT ) )
		{
			igems_FillField( game );
			dirty = 1;
			continue;
		}
	} // while

	if( dirty )
	{
		wid->dirty = 1;
	}
	return 0;
}


///////////////////////////////////////////////////////////
int igems_HandleEvents( PzEvent* ev )
{
	int retVal = TTK_EV_UNUSED;

	switch( ev->type )
	{
		case PZ_EVENT_SCROLL:
			if( !IGEMS_IS_GAME_BUSY( g_globals.game.flags ) )
			{
				TTK_SCROLLMOD( ev->arg, 5 );
				retVal = igems_ScrollCursor( &(g_globals.game), ev->arg );
			}
			break;

		case PZ_EVENT_BUTTON_UP:
			break;

		case PZ_EVENT_BUTTON_DOWN:
			switch( ev->arg )
			{
				case PZ_BUTTON_ACTION:
					DEBUG_LOG( "%s(): PZ_BUTTON_PLAY\n", __func__ );
					if( !IGEMS_IS_GAME_BUSY( g_globals.game.flags ) )
					{
						igems_SelectCursor( &(g_globals.game) );
						retVal = TTK_EV_CLICK;
					}
					break;

				case PZ_BUTTON_PLAY:
					DEBUG_LOG( "%s: PZ_BUTTON_PLAY\n", __func__ );
					if( !IGEMS_IS_GAME_BUSY( g_globals.game.flags ) )
					{
						g_globals.game.cursor.pressed = FALSE;
						retVal = TTK_EV_CLICK;
					}
					break;

				case PZ_BUTTON_NEXT:
					DEBUG_LOG( "%s: PZ_BUTTON_NEXT\n", __func__ );
					if( !IGEMS_IS_GAME_BUSY( g_globals.game.flags ) )
					{
						igems_ScrollCursor( &(g_globals.game), 1 );
						retVal = TTK_EV_CLICK;
					}
					break;

				case PZ_BUTTON_PREVIOUS:
					DEBUG_LOG( "%s: PZ_BUTTON_PREVIOUS\n", __func__ );
					if( !IGEMS_IS_GAME_BUSY( g_globals.game.flags ) )
					{
						igems_ScrollCursor( &(g_globals.game), -1 );
						retVal = TTK_EV_CLICK;
					}
					break;

				case PZ_BUTTON_HOLD:
					DEBUG_LOG( "%s: PZ_BUTTON_HOLD\n",  __func__ );
					break;

				case PZ_BUTTON_MENU:
					DEBUG_LOG( "%s: PZ_BUTTON_MENU\n", __func__ );
					pz_close_window( ev->wid->win );
					// call cleanup from here??
					break;
			}
			break;

		case PZ_EVENT_DESTROY:
			DEBUG_LOG( "%s: PZ_EVENT_DESTROY\n", __func__ );
			igems_Cleanup();
			break;
	}

	if( retVal != TTK_EV_UNUSED )
	{
		ev->wid->win->dirty = 1;
	}
	return retVal;
}

///////////////////////////////////////////////////////////
static int igems_SaveSettings( TWidget *this, int key, int time )
{
	if( key == TTK_BUTTON_MENU )
	{
		INFO_LOG( "Saving iGems settings ...\n" );
		pz_save_config( g_globals.config );
	}
	return ttk_menu_button( this, key, time );
}

///////////////////////////////////////////////////////////
/*! Called when the user selects our menu entry. */
PzWindow* igems_NewWindow( void )
{
	DEBUG_LOG( "%s()\n", __func__ );

	JFS_ASSERT( (g_globals.game.field == NULL) )

	// initialize the random number generator ...
	srand( time(NULL) );

	g_globals.game.field = igems_CreateField();

	// initialize the field and clear out any initial matches.
	igems_InitField( g_globals.game.field );
	while( igems_CheckForMatches( &(g_globals.game) ) )
	{
		igems_CleanField( &(g_globals.game) );
		igems_CollapseField( &(g_globals.game) );
		igems_FillField( &(g_globals.game) );
	}
	g_globals.game.flags = 0;

	igems_InitCursor( &(g_globals.game.cursor) );

	// TODO - initialize the renderer
	if( !igems_InitRenderer() )
	{
		ERROR_LOG( "Failed to initialize the renderer, cannot continue.\n" );
		igems_DeinitRenderer();
		return NULL;  // is NULL the right answer here?
	}

	// create a new window
	g_globals.window = pz_new_window( IGEMS_TITLE, PZ_WINDOW_NORMAL );

#if IGEMS_FULLSCREEN
	ttk_window_hide_header( g_globals.window );
	g_globals.window->x = 0;
	g_globals.window->y = 0;
	g_globals.window->w = ttk_screen->w;
	g_globals.window->h = ttk_screen->h;
#endif

	g_globals.widget = pz_add_widget(
			g_globals.window,
			igems_Render,
			igems_HandleEvents );

	pz_widget_set_timer( g_globals.widget, IGEMS_LOOP_TIME );
	g_globals.widget->timer = igems_GameLoop;

	return pz_finish_window( g_globals.window );
}

///////////////////////////////////////////////////////////
/*! Module entry point for initialization. */
void igems_Init( void )
{
	// TODO - once the black-white skin is ready, add it here as an option.
	static const char* skin_options[] = { "Gems", "Blocks", /* "B+W", */ NULL };
	static const char* cursor_options[] = { "Simple", "Fancy", NULL };

	DEBUG_LOG( "<<<<<<<<<<\n" );

	// initialize the game data and field ...
	memset( (void*) &(g_globals.game), 0, sizeof(IGEMS_GAMEDATA) );

	// register the module and our cleanup callback
	g_globals.module = pz_register_module( "igems", igems_Cleanup );

	// this stays allocated for the life of the program (i.e. do not try to free() in the cleanup function).
	g_globals.config = pz_load_config( pz_module_get_cfgpath(g_globals.module, "igems.conf") );

	// add our menu entry, and link it our main entry point
	pz_menu_add_action( IGEMS_MENU_PATH "/Play", igems_NewWindow );

	pz_menu_add_setting( IGEMS_MENU_PATH "/Settings/Skin", IGEMS_SKIN_OPTION, g_globals.config, skin_options );
	pz_menu_add_setting( IGEMS_MENU_PATH "/Settings/Cursor Type", IGEMS_CURSOR_OPTION, g_globals.config, cursor_options );
	

	// set callback so we know when to save our settings
	((TWidget *) pz_get_menu_item (IGEMS_MENU_PATH)->data)->button = igems_SaveSettings;

	DEBUG_LOG( "igems loaded, action set.\n" );
}

// Declare our module init function.  Should always be the last thing in the file
// Note - Do not use semicolon.
PZ_MOD_INIT( igems_Init )
