/*
 * Copyright (C) 2006 Jason Smith
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <pz.h>
#include <errno.h>
#include <string.h>

#include "defs.h"
#include "render.h"
#include "util.h"
#include "game.h"
#include "log.h"

/////////////////////////////////////////////////////////////////////

#define IGEMS_EVEN_FILE      "even.png"
#define IGEMS_ODD_FILE       "odd.png"
#define IGEMS_KILL_FILE      "kill2.png"

#define IGEMS_BLOCK_MARGIN  2

#define IGEMS_SKIN_TILED      0
#define IGEMS_SKIN_BLOCKS     1
#define IGEMS_SKIN_BW         2

#define IGEMS_CURSOR_SIMPLE   0
#define IGEMS_CURSOR_FANCY    1

/////////////////////////////////////////////////////////////////////

// define the different colors
enum BLAH
{
	IGEMS_COLOR_BG     = 0,

	IGEMS_SQUARE_GEM   = 1,
	IGEMS_TRIANGLE_GEM = 2,
	IGEMS_RSQUARE_GEM  = 3,
	IGEMS_DIAMOND_GEM  = 4,
	IGEMS_CIRCLE_GEM   = 5,
	IGEMS_HEXAGON_GEM  = 6,
	IGEMS_OCTAGON_GEM  = 7,
	
	IGEMS_COLOR_CURSOR,
	IGEMS_COLOR_KILL,

	// must be last
	IGEMS_NUM_COLORS   
};

////////////////////////////////////////////////////////

// holds our file-global data
static struct
{
	int           skinIndex;
	int           cursorType;

	// info for the tiled skin
	ttk_surface   evenImage;
	ttk_surface   oddImage;
	ttk_surface   killImage;

	int           needScaled;

	ttk_color     colors[ IGEMS_NUM_COLORS ];
} s_rdata;


/////////////////////////////////////////////////////////////////////
static int igems_LoadImage( ttk_surface* srf, const char* file )
{
	if( !(*srf = ttk_load_image( pz_module_get_datapath(g_globals.module, file) )) )
	{
		int myerrno = errno;
		ERROR_LOG( "Failed to load image '%s': %s\n", file, strerror(myerrno) );
		return FALSE;
	}

	return TRUE;
}

/////////////////////////////////////////////////////////////////////
static void igems_FreeImage( ttk_surface* srf )
{
	if( *srf )
	{
		ttk_free_surface( *srf );
		*srf = 0;
	}
}

/////////////////////////////////////////////////////////////////////
int igems_InitRenderer( void )
{
	// clear out our data structure
	memset( (void*) &s_rdata, 0, sizeof(s_rdata) );

	JFS_ASSERT( g_globals.config );
	s_rdata.skinIndex = pz_get_int_setting( g_globals.config, IGEMS_SKIN_OPTION );
	s_rdata.cursorType = pz_get_int_setting( g_globals.config, IGEMS_CURSOR_OPTION );

	// make sure the values are reasonable.
	if( s_rdata.skinIndex < 0 )
		s_rdata.skinIndex = 0;
	if( s_rdata.skinIndex > 2 )
		s_rdata.skinIndex = 2;

	if( s_rdata.cursorType < 0 )
		s_rdata.cursorType = 0;
	if( s_rdata.cursorType > 1 )
		s_rdata.cursorType = 1;

	// setup our colors
	s_rdata.colors[ IGEMS_COLOR_CURSOR ] = ttk_makecol( WHITE );
	s_rdata.colors[ IGEMS_COLOR_BG     ] = ttk_makecol( BLACK );
	s_rdata.colors[ IGEMS_COLOR_KILL   ] = ttk_makecol( 225, 225, 225 );

	s_rdata.colors[ IGEMS_SQUARE_GEM   ] = ttk_makecol( 255, 0, 0 );
	s_rdata.colors[ IGEMS_TRIANGLE_GEM ] = ttk_makecol( 255, 0, 255 );
	s_rdata.colors[ IGEMS_RSQUARE_GEM  ] = ttk_makecol( 255, 255, 0 );
	s_rdata.colors[ IGEMS_DIAMOND_GEM  ] = ttk_makecol( 128, 128, 255 );
	s_rdata.colors[ IGEMS_CIRCLE_GEM   ] = ttk_makecol( 225, 225, 245 );
	s_rdata.colors[ IGEMS_HEXAGON_GEM  ] = ttk_makecol( 255, 128, 0 );
	s_rdata.colors[ IGEMS_OCTAGON_GEM  ] = ttk_makecol( 0, 255, 0 );

	// indicate that the images need scaled down to our display
	s_rdata.needScaled = TRUE;

	switch( s_rdata.skinIndex )
	{
		case IGEMS_SKIN_TILED:
			// load the images
			if( igems_LoadImage( &(s_rdata.oddImage),  IGEMS_ODD_FILE  ) &&
				igems_LoadImage( &(s_rdata.evenImage), IGEMS_EVEN_FILE ) &&
				igems_LoadImage( &(s_rdata.killImage), IGEMS_KILL_FILE ) )
			{
				DEBUG_LOG( "Successfully loaded all images.\n" );
			}
			else
			{
				ERROR_LOG( "Failed to load one or more images,... %s() failed.\n", __func__ );
				return FALSE;
			}
			break;

		case IGEMS_SKIN_BLOCKS:
			// do nothing
			break;

		case IGEMS_SKIN_BW:
			// do nothing
			break;

		default:
			ERROR_LOG( "Invalid skin index %i\n", s_rdata.skinIndex );
			JFS_ASSERT( 0 );
			return FALSE;
	}
	
	return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
void igems_DeinitRenderer( void )
{
	// free our images (if loaded)
	igems_FreeImage( &(s_rdata.oddImage) );
	igems_FreeImage( &(s_rdata.evenImage) );
	igems_FreeImage( &(s_rdata.killImage) );
}

///////////////////////////////////////////////////////////////////////////////
static void igems_DrawField_Tiled( ttk_surface srf, int size, int fx, int fy )
{
	int i, x, y;
	int sx, sy;
	IGEMS_BLOCK* b;
	ttk_surface tmp;

	// check the dimensions of our source images
	if( s_rdata.needScaled )
	{
		int w, h;
		ttk_surface_get_dimen( s_rdata.oddImage, &w, &h );
		if( (size*3) != w )
		{
			float scale = ((float)size) * 3.0f / ((float)w);

			tmp = ttk_scale_surface( s_rdata.evenImage, scale );
			ttk_free_image( s_rdata.evenImage );
			s_rdata.evenImage = tmp;

			tmp = ttk_scale_surface( s_rdata.oddImage, scale );
			ttk_free_image( s_rdata.oddImage );
			s_rdata.oddImage = tmp;

			tmp = ttk_scale_surface( s_rdata.killImage, scale );
			ttk_free_image( s_rdata.killImage );
			s_rdata.killImage = tmp;
		}

		// no longer the first time here
		s_rdata.needScaled = FALSE;
	} // if need to scale

#if 0
	// here is the code for only drawing dirty columns,
	// maybe when hotdog comes out we'll actually be able to use it.
	for( x=0; x<IGEMS_FIELD_SIZE; ++x )
	{
		if( IS_COL_DIRTY( x ) )
		{
			for( y=0; y<IGEMS_FIELD_SIZE; ++y )
			{
				b = igems_GetBlock( g_globals.game.field, x, y );
				JFS_ASSERT( b );

				if( b->type != IGEMS_NO_BLOCK )
				{
					sx = ((b->type - 1) % 3) * size;
					sy = ((b->type - 1) / 3) * size;

					if( b->kill )
						tmp = s_rdata.killImage;
					else if( (x+y) & 0x1 )
						tmp = s_rdata.oddImage;
					else
						tmp = s_rdata.evenImage;
			
					ttk_blit_image_ex( tmp, sx, sy, size, size, srf, (x*size+fx), (y*size+fy) );

				} // if valid block
				else
				{
					// clear the space by drawing the background
					ttk_fillrect( srf,
						x*size+fx,      y*size+fy,
						x*size+fx+size, y*size+fy+size,
						s_rdata.colors[ IGEMS_COLOR_BG ] );
				}
			} // for every row
		}
	}
	g_globals.game.dirtyCols = 0;  // clear all dirty columns

#else
	// draw all blocks in the field
	for( i=0; i<(IGEMS_FIELD_SIZE*IGEMS_FIELD_SIZE); i++ )
	{
		x = i % IGEMS_FIELD_SIZE;
		y = i / IGEMS_FIELD_SIZE;

		b = igems_GetBlock( g_globals.game.field, x, y );
		JFS_ASSERT( b );

		if( b->type != IGEMS_NO_BLOCK )
		{
			sx = ((b->type - 1) % 3) * size;
			sy = ((b->type - 1) / 3) * size;

			if( b->kill )
				tmp = s_rdata.killImage;
			else if( (x+y) & 0x1 )
				tmp = s_rdata.oddImage;
			else
				tmp = s_rdata.evenImage;
			
			ttk_blit_image_ex( tmp, sx, sy, size, size, srf, (x*size+fx), (y*size+fy) );

		} // if valid block
	} // for every block

#endif
}

///////////////////////////////////////////////////////////////////////////////
static void igems_DrawField_Simple( ttk_surface srf, int size, int fx, int fy )
{
	int x, y;
	IGEMS_BLOCK* b;

	// draw all blocks in the field
	for( y=0; y<IGEMS_FIELD_SIZE; ++y )
	{
		for( x=0; x<IGEMS_FIELD_SIZE; ++x )
		{
			b = igems_GetBlock( g_globals.game.field, x, y );
			JFS_ASSERT( b );

			//if( b->type != IGEMS_NO_BLOCK )
			{
				if( b->kill )
				{
					// draw something to show it's being killed
					ttk_fillrect( srf,
							(x*size + fx), (y*size + fy),
							(x*size + fx + size), (y*size + size + fy),
							s_rdata.colors[ IGEMS_COLOR_KILL ] );
				}

				// draw the actual block.
				ttk_fillrect( srf,
						(x*size + fx + IGEMS_BLOCK_MARGIN),        (y*size + fy + IGEMS_BLOCK_MARGIN),
						(x*size + fx - IGEMS_BLOCK_MARGIN + size), (y*size + fy - IGEMS_BLOCK_MARGIN + size),
						s_rdata.colors[ b->type ] );
			} // if valid block
		} // for every x
	} // for every y
}

///////////////////////////////////////////////////////////////////////////////
static void igems_DrawCursorGuts( ttk_surface srf, int x, int y, int size, int dir )
{
	ttk_color col = s_rdata.colors[ IGEMS_COLOR_CURSOR ];

	// top
	if( dir != IGEMS_CURSOR_UP )
	{
		ttk_fillrect( srf, 
				x, y, 
				x+size, y+IGEMS_CURSOR_WIDTH, 
				col );
	}

	// right
	if( dir != IGEMS_CURSOR_RIGHT )
	{
		ttk_fillrect( srf,
				x+size-IGEMS_CURSOR_WIDTH, y,
				x+size, y+size,
				col );
	}

	// bottom
	if( dir != IGEMS_CURSOR_DOWN )
	{
		ttk_fillrect( srf,
				x, y+size-IGEMS_CURSOR_WIDTH,
				x+size, y+size,
				col );
	}

	// left
	if( dir != IGEMS_CURSOR_LEFT )
	{
		ttk_fillrect( srf,
				x, y,
				x+IGEMS_CURSOR_WIDTH, y+size,
				col );
	}
}

///////////////////////////////////////////////////////////////////////////////
static void igems_DrawFancyCursorGuts( ttk_surface srf, int x, int y, int size, int dir )
{
	ttk_color col = s_rdata.colors[ IGEMS_COLOR_CURSOR ];
	int size3  =   size/3;
	int size23 = 2*size/3;

	// top
	if( dir != IGEMS_CURSOR_UP )
	{
		ttk_fillrect( srf, 
				x, y, 
				x+size3, y+IGEMS_CURSOR_WIDTH, 
				col );

		ttk_fillrect( srf, 
				x+size23, y, 
				x+size, y+IGEMS_CURSOR_WIDTH, 
				col );
	}

	// right
	if( dir != IGEMS_CURSOR_RIGHT )
	{
		ttk_fillrect( srf,
				x+size-IGEMS_CURSOR_WIDTH, y,
				x+size, y+size3,
				col );

		ttk_fillrect( srf,
				x+size-IGEMS_CURSOR_WIDTH, y+size23,
				x+size, y+size,
				col );
	}

	// bottom
	if( dir != IGEMS_CURSOR_DOWN )
	{
		ttk_fillrect( srf,
				x, y+size-IGEMS_CURSOR_WIDTH,
				x+size3, y+size,
				col );

		ttk_fillrect( srf,
				x+size23, y+size-IGEMS_CURSOR_WIDTH,
				x+size, y+size,
				col );
	}

	// left
	if( dir != IGEMS_CURSOR_LEFT )
	{
		ttk_fillrect( srf,
				x, y,
				x+IGEMS_CURSOR_WIDTH, y+size3,
				col );

		ttk_fillrect( srf,
				x, y+size23,
				x+IGEMS_CURSOR_WIDTH, y+size,
				col );
	}
}


///////////////////////////////////////////////////////////////////////////////
/*! Draw the game's cursor at the specified location. */
static void igems_DrawCursor( ttk_surface srf, int x, int y, int blockSize )
{
	void (*draw_cursor_func)( ttk_surface, int, int, int, int ) = igems_DrawCursorGuts;

	if( s_rdata.cursorType == IGEMS_CURSOR_FANCY )
		draw_cursor_func = igems_DrawFancyCursorGuts;

	if( g_globals.game.cursor.pressed )
	{
		int dir = g_globals.game.cursor.direction;

		// draw the actual cursor
		draw_cursor_func( srf, x, y, blockSize, dir );

		// get info for the selected swap block
		switch( g_globals.game.cursor.direction )
		{
			case IGEMS_CURSOR_UP:     y-=blockSize;  dir = IGEMS_CURSOR_DOWN;  break;
			case IGEMS_CURSOR_RIGHT:  x+=blockSize;  dir = IGEMS_CURSOR_LEFT;  break;
			case IGEMS_CURSOR_DOWN:   y+=blockSize;  dir = IGEMS_CURSOR_UP;    break;
			case IGEMS_CURSOR_LEFT:   x-=blockSize;  dir = IGEMS_CURSOR_RIGHT; break;
		}

		// actually draw the swap block
		draw_cursor_func( srf, x, y, blockSize, dir );
	}
	else // cursor not pressed, so just draw lines all around the block.
	{
		draw_cursor_func( srf, x, y, blockSize, (-1) );
	}
}

///////////////////////////////////////////////////////////////////////////////
void igems_Render( PzWidget* wid, ttk_surface srf )
{
	int blockSize = ( wid->h / (IGEMS_FIELD_SIZE) );
	int offset = (wid->w - (IGEMS_FIELD_SIZE * blockSize)) - blockSize/2;

	ttk_fillrect( srf, 0, 0, wid->w, wid->h,  s_rdata.colors[ IGEMS_COLOR_BG ] );

	// draw the entire game field
	switch( s_rdata.skinIndex )
	{
		case IGEMS_SKIN_TILED:
			igems_DrawField_Tiled( srf, blockSize, offset, 0 );
			break;

		case IGEMS_SKIN_BW:
			// todo

		case IGEMS_SKIN_BLOCKS:
		default:
			igems_DrawField_Simple( srf, blockSize, offset, 0 );
			break;
	} // switch

	// draw the cursor, but only if it isn't busy with an animation
	if( !IGEMS_IS_GAME_BUSY(g_globals.game.flags) )
	{
		igems_DrawCursor( srf,
				g_globals.game.cursor.x * blockSize + offset,
				g_globals.game.cursor.y * blockSize,
				blockSize );
	}

	// draw other elements if needed (score, time, etc).
	// this is just a demo, so don't bother with anything else right now

// 	ttk_draw_window( wid->win );
// 	ttk_gfx_update( ttk_screen->srf );
}

